<?php

declare(strict_types=1);

namespace Drupal\geofield_maplibre_gl\Plugin\Field\FieldWidget;

use Drupal\Core\Field\FieldItemListInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\geofield\Plugin\Field\FieldWidget\GeofieldLatLonWidget;

/**
 * Defines the 'geofield_maplibre_gl' field widget.
 *
 * @FieldWidget(
 *   id = "geofield_maplibre_gl",
 *   label = @Translation("Geofield Maplibre"),
 *   field_types = {"geofield"},
 * )
 */
final class GeofieldMaplibreGlWidget extends GeofieldLatLonWidget implements ContainerFactoryPluginInterface {

  /**
   * {@inheritdoc}
   */
  public static function defaultSettings(): array {
    $setting = [
      'style' => 'https://demotiles.maplibre.org/style.json',
    ];
    return $setting + parent::defaultSettings();
  }

  /**
   * {@inheritdoc}
   */
  public function settingsForm(array $form, FormStateInterface $form_state): array {
    $elements = parent::settingsForm($form, $form_state);
    $elements['style'] = [
      '#title' => $this->t('Style'),
      '#type' => 'textfield',
      '#default_value' => $this->getSetting('style'),
      '#required' => TRUE,
    ];

    return $elements;
  }

  /**
   * {@inheritdoc}
   */
  public function settingsSummary(): array {
    $summary = parent::settingsSummary();
    $summary[] = $this->t('Style: @style', ['@style' => $this->getSetting('style')]);
    return $summary;
  }

  /**
   * {@inheritdoc}
   */
  public function formElement(FieldItemListInterface $items, $delta, array $element, array &$form, FormStateInterface $form_state): array {
    $element = parent::formElement($items, $delta, $element, $form, $form_state);
    $map_id = "maplibre-gl-widget-{$items->getName()}-{$delta}";
    /** @var \Drupal\geofield\Plugin\Field\FieldType\GeofieldItem $geofield_item */
    $geofield_item = $items->getValue()[$delta];
    if (empty($geofield_item) || $geofield_item['geo_type'] == 'Point') {
      $latlon_value = [];

      foreach ($this->components as $component) {
        $latlon_value[$component] = isset($items[$delta]->{$component}) ? floatval($items[$delta]->{$component}) : '';
      }
      $element['value'] += [
        '#type' => 'geofield_latlon',
        '#default_value' => $latlon_value,
        '#geolocation' => $this->getSetting('html5_geolocation'),
        '#error_label' => !empty($element['#title']) ? $element['#title'] : $this->fieldDefinition->getLabel(),
      ];
      $map_info = [
        // 'access_token' => $this->configuration['mapbox_api_key'],
        'options' => [
          'container' => $map_id,
          'style' => $this->getSetting('style'),
          'zoom' => 10,
          'center' => [
            (isset($this->getSetting("center")["lon"])) ? $this->getSetting("center")["lon"] : 0,
            (isset($this->getSetting("center")["lat"])) ? $this->getSetting("center")["lat"] : 0,
          ],
        ],
        'config' =>
        [
          'controls' => [],
        ],
      ];

      $element['value']['maplibre-container'] = [
        '#prefix' => '<div class="maplibre-gl-wrapper maplibre-widget">',
        '#markup' => '<div id="' . $map_id . '-menu" class="maplibre-gl-layer-menu"></div>'
        . '<div id="' . $map_id . '" class="maplibre-gl-container"></div>',
        '#suffix' => '</div">',
      ];

      $element['#attached']['drupalSettings']['maplibreGl'] = [$map_id => $map_info];
      $element['#attached']['drupalSettings']['geofieldMaplibreGlFieldWidget'][$map_id] = [
        'pickerType' => 'widget',
        'fieldName' => $items->getName(),
        'delta' => $delta,
      ];

      $element['#attached']['library'][] = 'geofield_maplibre_gl/field_widget.maplibre_gl';
      $element['#attached']['library'][] = 'maplibre_gl/libraries.maplibre-gl-js';
      $element['#attached']['library'][] = 'maplibre_gl/maplibre_gl';

    }
    else {
      $widget_label = $this->getPluginDefinition()['label']->render();
      $element['value'] += [
        '#prefix' => '<div class="geofield-warning">' . $this->t('The "@widget_label" widget cannot be applied because it doesn\'t support Geometries (Polylines, Polygons, etc.).', [
          '@widget_label' => $widget_label,
        ]) . '</div>',
        '#type' => 'textarea',
        '#default_value' => $items[$delta]->value ?: NULL,
      ];
    }
    return $element;
  }

}
